(* $Id: taskalg.ml,v 1.60 2011-10-28 13:52:31 deraugla Exp $ *)

open Ast;
open Printf;

type sarray_t 'a = Sarray.t 'a;
value sarray_empty = Sarray.empty;

type stask =
  { st_release : int;
    st_init : sarray_t (option time_unit);
    st_repeat : sarray_t (option time_unit);
    st_contents : list (int * Ast.task_def) }
and time_unit =
  { tu_itask : int;
    tu_is_preemption : bool;
    tu_is_first : bool }
;

type stask_bucket =
  { sb_id : int;
    sb_task : stask;
    sb_pos : int;
    sb_len : int;
    sb_inst_beg : int;
    sb_next_deadline : int;
    sb_preemption : int }
;

value long_display = ref False;

value string_of_sarray string_of_elem arr =
  let len = Sarray.length arr in
  if not long_display.val && len > 300 then do {
    let slen = 30 in
    let sarr1 = Sarray.sub arr 0 slen in
    let sarr2 = Sarray.sub arr (len - slen) slen in
    let list1 = List.map string_of_elem (Sarray.to_list sarr1) in
    let list2 = List.map string_of_elem (Sarray.to_list sarr2) in
    let s1 = String.concat "" list1 in
    let s2 = String.concat "" list2 in
    sprintf "%s...(%d tu)...%s" s1 (len - 2 * slen) s2
  }
  else do {
    let list = Sarray.to_list arr in
    let list = List.rev_map string_of_elem (List.rev list) in
    String.concat "" list
  }
;

value col = [| 2; 1; 4; 5; 6; 3 |];
value col tu = do {
  let n = tu.tu_itask in
  if n < 0 then 0 else col.(n mod Array.length col)
};

value string_of_option_unit colour =
  fun
  [ Some tu -> do {
      if colour then do {
        sprintf "\027[%sm%d\027[m"
          (if tu.tu_is_preemption then "7" else sprintf "1;3%d" (col tu))
          (tu.tu_itask + 1)
      }
      else do {
        sprintf "%s%d%s" (if tu.tu_is_preemption then "(" else "")
          (tu.tu_itask + 1) (if tu.tu_is_preemption then ")" else "")
      }
    }
  | None -> "." ]
;

value string_of_stask_area colour =
  string_of_sarray (string_of_option_unit colour)
;

value to_string colour st = do {
  sprintf "%s{%s} (rel %d)" (string_of_stask_area colour st.st_init)
    (string_of_stask_area colour st.st_repeat) st.st_release
};

value repeat_of_task_def i td =
  loop [] td.duration td.period
  where rec loop rev_list dur per =
    if per = 0 then Sarray.of_list (List.rev rev_list)
    else if dur = 0 then loop [None :: rev_list] 0 (per - 1)
    else do {
      let tu =
        let is_first = rev_list = [] in
        {tu_itask = i; tu_is_preemption = False; tu_is_first = is_first}
      in
      loop [Some tu :: rev_list] (dur - 1) (per - 1)
    }
;

value of_task_def i td = do {
  if td.sub_duration <> [] then
    failwith (sprintf "sub duration not impl (t%d)" (i + 1))
  else ();
  {st_release = td.release; st_init = sarray_empty;
   st_repeat = repeat_of_task_def i td; st_contents = [(i, td)] }
};

value new_id = do {
  let cnt = ref 0 in
  fun () -> do {
    incr cnt;
    cnt.val
  }
};

value td_of_stask st =
  match st.st_contents with
  [ [(_, td) :: _] -> td
  | [] -> failwith "td_of_stask" ]
;

value stask_bucket_of_stask st =
  let td = td_of_stask st in
  {sb_id = new_id (); sb_task = st; sb_pos = 0;
   sb_len = Sarray.length st.st_init; sb_inst_beg = st.st_release;
   sb_next_deadline = st.st_release + td.deadline; sb_preemption = 0}
;

value stask_next sb = do {
  if sb.sb_pos >= sb.sb_len then
    None
  else
    let v = Sarray.get sb.sb_task.st_init sb.sb_pos in
    let sb = {(sb) with sb_pos = sb.sb_pos + 1} in
    Some (v, sb)
};

value cnt = ref 0;
value stask_bucket_next sb = do {
let _ = incr cnt in
  let st = sb.sb_task in
  if sb.sb_pos < sb.sb_len then do {
    let v = Sarray.get st.st_init sb.sb_pos in
    let sb = {(sb) with sb_pos = sb.sb_pos + 1} in
    (v, sb)
  }
  else do {
    let st = {(st) with st_init = st.st_repeat} in
    let sb =
      {(sb) with sb_task = st; sb_pos = 1;
       sb_len = Sarray.length st.st_init}
    in
    (Sarray.get st.st_init 0, sb)
  }
};

value update_release tasks_prec stask1 stask2 (itask2, td2) = do {
  match tasks_prec with
  [ [] -> Left stask2
  | pt -> do {
      let release_opt =
        loop_prec stask2.st_release pt where rec loop_prec release =
          fun
          [ [itask :: itaskl] -> do {
               loop stask1.st_contents where rec loop =
                 fun
                 [ [(itask1, td1) :: tdl] -> do {
                     if itask1 = itask then do {
                       let d = td1.duration in
                       let sb = stask_bucket_of_stask stask1 in
                       loop stask1.st_release d (stask_bucket_next sb)
                       where rec loop rel d (v, sb) =
                         if d = 0 then loop_prec (max rel release) itaskl
                         else do {
                           let d =
                             match v with
                             [ Some tu ->
                                 if tu.tu_itask = itask &&
                                    not tu.tu_is_preemption
                                 then
                                   d - 1
                                 else d
                             | None -> d ]
                           in
                           loop (rel + 1) d (stask_bucket_next sb)
                         }
                     }
                     else loop tdl
                   }
                 | [] -> Right "glop" ]
            }
          | [] -> Left release ]
      in
      match release_opt with
      [ Left release -> do {
          let td2 = {(td2) with release = release} in
          let contents = [(itask2, td2)] in
          Left {(stask2) with st_release = release; st_contents = contents}
        }
      | Right msg -> Right msg ]
    } ]
};

value adjust_release diff_rel sb =
  let st =
    let st = sb.sb_task in
    {(st) with
     st_release = st.st_release - diff_rel;
     st_init = Sarray.append (Sarray.make diff_rel None) st.st_init}
  in
  {sb_id = sb.sb_id; sb_task = st; sb_pos = 0;
   sb_len = Sarray.length st.st_init; sb_inst_beg = sb.sb_inst_beg;
   sb_next_deadline = sb.sb_next_deadline; sb_preemption = sb.sb_preemption}
;

value adjust_releases sb1 sb2 =
  let st1 = sb1.sb_task in
  let st2 = sb2.sb_task in
  if st1.st_release < st2.st_release then do {
    let n = st2.st_release - st1.st_release in
    let sb2 = adjust_release n sb2 in
    (sb1, sb2)
  }
  else if st1.st_release > st2.st_release then do {
    let n = st1.st_release - st2.st_release in
    let sb1 = adjust_release n sb1 in
    (sb1, sb2)
  }
  else (sb1, sb2)
;

module type S =
  sig
    type t 'a = 'abstract;
    value empty : t 'a;
    value length : t 'a -> int;
    value last : t 'a -> option 'a;
    value add : 'a -> t 'a -> t 'a;
    value get_init_and_repeat :
      int -> t 'a -> option (sarray_t 'a * sarray_t 'a);
    value to_string : ('a -> string) -> t 'a -> string;
  end
;

module S1 : S =
  struct
    type t 'a = list 'a;
    value empty = [];
    value length = List.length;
    value last =
      fun
      [ [x :: _] -> Some x
      | [] -> None ]
    ;
    value add x t = [x :: t];
    value list_rev_first_elem n list =
      loop n [] list where rec loop n r list =
        if n = 0 then (r, list)
        else do {
          match list with
          [ [x :: l] -> loop (n - 1) [x :: r] l
          | [] -> (r, []) ]
        }
    ;
    value get_init_and_repeat n t = do {
      let (list1, rev_rest) = list_rev_first_elem n t in
      let (list2, rev_rest) = list_rev_first_elem n rev_rest in
      if list1 = list2 then do {
        let (rev_init, repeat_hd, rev_repeat) =
          loop rev_rest [] (List.rev list1)
          where rec loop rev_init repeat_hd rev_repeat =
            match (rev_init, rev_repeat) with
            [ ([x :: l1], [y :: l2]) -> do {
                if x = y then loop l1 [y :: repeat_hd] l2
                else (rev_init, repeat_hd, rev_repeat)
              }
            | (_, _) -> (rev_init, repeat_hd, rev_repeat) ]
        in
        let init = Sarray.of_list (List.rev rev_init) in
        let repeat = Sarray.of_list (repeat_hd @ (List.rev rev_repeat)) in
        Some (init, repeat)
      }
      else None
    };
    value to_string string_of_elem list =
      let list = List.rev_map string_of_elem list in
      String.concat "" list
    ;
  end
;

module S2 : S =
  struct
    type t 'a = {ea_arr : sarray_t 'a; ea_pos : int};
    value empty = {ea_arr = sarray_empty; ea_pos = 0};
    value length ea = ea.ea_pos;
    value last ea = do {
      let pos = ea.ea_pos in
      if pos > 0 then Some (Sarray.get ea.ea_arr (pos - 1)) else None
    };
    value add x ea = do {
      let len = Sarray.length ea.ea_arr in
      let a =
        if ea.ea_pos = len then do {
          let new_len = 2 * len + 1 in
          Sarray.init new_len
            (fun i -> if i < len then Sarray.get ea.ea_arr i else x)
        }
        else do {
          Sarray.set ea.ea_arr ea.ea_pos x;
          ea.ea_arr
        }
      in
      {ea_arr = a; ea_pos = ea.ea_pos + 1}
    };
    value get_init_and_repeat n ea = do {
      if ea.ea_pos < n then None
      else do {
        let init_len =
          loop n (ea.ea_pos - 1) (ea.ea_pos - n - 1) where rec loop k i j =
            if j < 0 then
              if k <= 0 then 0 else -1
            else if Sarray.get ea.ea_arr i = Sarray.get ea.ea_arr j then
              loop (k - 1) (i - 1) (j - 1)
            else if k < 0 then j + 1
            else -1
        in
        if init_len >= 0 && init_len + n <= Sarray.length ea.ea_arr then do {
          let init = Sarray.sub ea.ea_arr 0 init_len in
          let repeat = Sarray.sub ea.ea_arr init_len n in
          Some (init, repeat)
        }
        else None
      }
    };
    value to_string string_of_elem ea =
      let a = Sarray.sub ea.ea_arr 0 ea.ea_pos in
      string_of_sarray string_of_elem a
(*
      let list = List.map string_of_elem (Sarray.to_list a) in
      String.concat "" list
*)
    ;
  end
;

module S = S2;

value check_repeat per_lcm accu sb1 sb2 = do {
  match S.get_init_and_repeat per_lcm accu with
  [ Some (init, repeat) -> do {
      let contents = sb1.sb_task.st_contents @ sb2.sb_task.st_contents in
      let st =
        {(sb1.sb_task) with
         st_init = init; st_repeat = repeat; st_contents = contents}
      in
      Some st
    }
  | None -> None ]
};

value string_of_td_list tdl = do {
  let sl = List.map (fun (i, _) -> sprintf "t%d" (i + 1)) tdl in
  String.concat "," sl
};

value shorten_end_if_free sb = do {
  if sb.sb_len = 0 then Right "internal error"
  else do {
    match Sarray.get sb.sb_task.st_init (sb.sb_len - 1) with
    [ None -> do {
        let sb = {(sb) with sb_pos = sb.sb_pos - 1; sb_len = sb.sb_len - 1} in
        Left sb
      }
    | Some _ -> do {
        let s = string_of_td_list sb.sb_task.st_contents in
        Right (sprintf "period of %s overstepped" s)
      } ]
  }
};

value preemption_cost_of_stask tc sb =
  let td = td_of_stask sb.sb_task in
  let pree_cost = td.pree_cost in
  if pree_cost = 0 then tc.preemption_cost
  else pree_cost
;

value select_unit tcon preemption v1 v2 sb1 sb2 = do {
  match (v1, v2) with
  [ (Some _, Some tu2) -> do {
      if sb2.sb_pos >= sb2.sb_len then do {
        let s = string_of_td_list sb2.sb_task.st_contents in
        Right (sprintf "period of %s overstepped" s)
      }
      else do {
        match shorten_end_if_free sb2 with
        [ Left sb2 -> do {
            let preemption =
              if not tu2.tu_is_first then preemption_cost_of_stask tcon sb2
              else 0
            in
            Left (sb1, sb2, v1, preemption)
          }
        | Right msg -> Right msg ]
      }
    }
  | (None, Some tu2) -> do {
      let td2_non_preemptive = (td_of_stask sb2.sb_task).non_preemptive in
      if td2_non_preemptive && tu2.tu_is_first then do {
        (* first time unit in non preemptive task: check if room enough;
         * put nothing (= None) if not. *)
        let dur =
          loop 1 sb2 where rec loop dur sb2 =
            match stask_next sb2 with
            [ Some (Some tu, sb2) -> do {
                if tu.tu_itask = tu2.tu_itask then loop (dur + 1) sb2
                else dur
              }
            | Some (None, _) | None -> dur ]
        in
        loop (dur - 1) sb1 where rec loop dur_rest sb =
          if dur_rest = 0 then Left (sb1, sb2, v2, 0)
          else do {
            let (v, sb) = stask_bucket_next sb in
            match v with
            [ Some _ -> do {
                match shorten_end_if_free sb2 with
                [ Left sb2 -> Left (sb1, sb2, None, 0)
                | Right msg -> Right msg ]
              }
            | None -> loop (dur_rest - 1) sb ]
          }
      }
      else if preemption > 0 then do {
        match Sarray.get sb2.sb_task.st_init (sb2.sb_len - 1) with
        [ None -> do {
            let sb2 =
              {(sb2) with sb_pos = sb2.sb_pos - 1; sb_len = sb2.sb_len - 1}
            in
            let tu =
              {tu_itask = tu2.tu_itask; tu_is_preemption = True;
               tu_is_first = False}
            in
            Left (sb1, sb2, Some tu, preemption - 1)
          }
        | Some _ -> Right "no room" ]
      }
      else Left (sb1, sb2, v2, 0)
    }
  | (Some _, None) -> Left (sb1, sb2, v1, 0)
  | (None, None) -> Left (sb1, sb2, None, 0) ]
};

(* algorithm of operation 'plus' *)

value plus_loop tcon per_lcm = do {
  loop S.empty 0 where rec loop accu p n sb1 sb2 = do {
    match if n = 0 then check_repeat per_lcm accu sb1 sb2 else None with
    [ Some r -> Left r
    | None -> do {
        let n = if n = 0 then per_lcm else n in
        let (v1, sb1) = stask_bucket_next sb1 in
        let (v2, sb2) = stask_bucket_next sb2 in
        match select_unit tcon p v1 v2 sb1 sb2 with
        [ Left (sb1, sb2, v, p) -> loop (S.add v accu) p (n - 1) sb1 sb2
        | Right msg -> Right msg ]
      } ]
  }
};

value plus tcon st1 st2 = do {
  let st2_opt =
    match st2.st_contents with
    [ [(itask2, td2)] -> do {
        let tasks_prec =
          Sparse.graph_preceding_tasks tcon.graph_precedence itask2
        in
        let tasks_prec =
          if tasks_prec = [] && tcon.global_precedence then do {
            List.map (fun (itask1, td1) -> itask1) st1.st_contents
          }
          else tasks_prec
        in
        update_release tasks_prec st1 st2 (itask2, td2)
      }
    | [] | [_ :: _] -> Left st2 ]
  in
  match st2_opt with
  [ Left st2 -> do {
      let per_lcm =
        Sutil.lcm (Sarray.length st1.st_repeat) (Sarray.length st2.st_repeat)
      in
      let sb1 = stask_bucket_of_stask st1 in
      let sb2 = stask_bucket_of_stask st2 in
      let (sb1, sb2) = adjust_releases sb1 sb2 in
      let rmax =
        max (Sarray.length st1.st_init) (Sarray.length st2.st_init)
      in
      let min_sched = rmax + 2 * per_lcm in
      plus_loop tcon per_lcm min_sched sb1 sb2
    }
  | Right msg -> Right msg ]
};

(* algorithm for EDF: all tasks together *)

value select_same_deadlines tc time accu =
  fun
  [ [] -> failwith "internal error"
  | [sb] -> sb
  | sbl -> do {
      let sb_opt =
        (* choose the one which was still running *)
        match S.last accu with
        [ Some (Some {tu_itask = itask}) -> do {
            loop sbl where rec loop =
              fun
              [ [sb :: sbl] -> do {
                  if List.mem_assoc itask sb.sb_task.st_contents then Some sb
                  else loop sbl
                }
              | [] -> None ]
          }
        | Some None | None -> None ]
      in
      match sb_opt with
      [ Some sb -> sb
      | None -> do {
          let (sbl1, cost) =
            (* choose the one which has the biggest preemption cost *)
            List.fold_left
              (fun (sbl, cost) sb ->
                 let pree_cost = preemption_cost_of_stask tc sb in
                 if pree_cost > cost then ([sb], pree_cost)
                 else if pree_cost < cost then (sbl, cost)
                 else ([sb :: sbl], cost))
              ([], 0) sbl
          in
if List.length sbl1 <> List.length sbl then failwith (sprintf "time %d worst preemption cost %d" time cost) else
          List.hd sbl1
        } ]
    } ]
;

value rec edf_apply_cand time tc cand_sb accu sbl =
  fun
  [ [(v, sb) :: rev_vsbl] -> do {
      if sb.sb_id = cand_sb.sb_id then do {
        if sb.sb_preemption > 0 then do {
          let tu =
            match v with
            [ Some tu -> {(tu) with tu_is_preemption = True}
            | None -> failwith "internal error" ]
          in
          let sb = {(sb) with sb_preemption = sb.sb_preemption - 1} in
          match shorten_end_if_free sb with
          [ Left sb ->
              edf_apply_cand time tc cand_sb (S.add (Some tu) accu)
                [sb :: sbl] rev_vsbl
          | Right msg ->
              Right (sprintf "time %d: %s" time msg) ]
        }
        else
          edf_apply_cand time tc cand_sb (S.add v accu) [sb :: sbl] rev_vsbl
      }
      else do {
        match v with
        [ Some tu -> do {
            match shorten_end_if_free sb with
            [ Left sb -> do {
                if tu.tu_is_first then
                  edf_apply_cand time tc cand_sb accu [sb :: sbl] rev_vsbl
                else do {
                  let pree_cost =
                    (td_of_stask sb.sb_task).pree_cost
                  in
                  let preemption =
                    if pree_cost = 0 then tc.preemption_cost else pree_cost
                  in
                  let sb = {(sb) with sb_preemption = preemption} in
                  edf_apply_cand time tc cand_sb accu [sb :: sbl] rev_vsbl
                }
              }
            | Right msg -> Right (msg ^ " (preemption)") ]
          }
        | None -> edf_apply_cand time tc cand_sb accu [sb :: sbl] rev_vsbl ]
      }
    }
  | [] -> Left (accu, sbl) ]
;
 
value rec edf_loop tc lcm time accu sbl = do {
  let sbl =
    List.map
      (fun sb ->
         if time >= sb.sb_next_deadline then do {
           let td = td_of_stask sb.sb_task in
           let next_inst = sb.sb_inst_beg + td.period in
           {(sb) with sb_inst_beg = next_inst;
            sb_next_deadline = next_inst + td.deadline}
         }
         else sb)
      sbl
  in
  let ((cand_sbl, deadline), rev_vsbl) =
    List.fold_left
      (fun ((cand_sbl, next_deadline), rev_vsbl) sb ->
         let st = sb.sb_task in
         if time >= st.st_release then do {
           let (v, sb) = stask_bucket_next sb in
           let cand_sbl_deadline =
             match v with
             [ Some v -> do {
                 let deadline = sb.sb_next_deadline in
                 if deadline < next_deadline then
                   ([sb], deadline)
                 else if deadline > next_deadline then
                   (cand_sbl, next_deadline)
                 else
                   ([sb :: cand_sbl], next_deadline)
               }
             | None -> (cand_sbl, next_deadline) ]
           in
           (cand_sbl_deadline, [(v, sb) :: rev_vsbl])
         }
         else ((cand_sbl, next_deadline), [(None, sb) :: rev_vsbl]))
      (([], max_int), []) sbl
  in
  let accu_sbl_opt =
    match cand_sbl with
    [ [] -> Left (S.add None accu, List.rev_map snd rev_vsbl)
    | _ -> do {
        let cand_sb = select_same_deadlines tc time accu cand_sbl in
        edf_apply_cand time tc cand_sb accu [] rev_vsbl
      } ]
  in
  match accu_sbl_opt with
  [ Left (accu, sbl) -> do {
      if S.length accu mod lcm <> 0 then edf_loop tc lcm (time + 1) accu sbl
      else do {
        loop lcm where rec loop n =
          if S.length accu >= 2 * n then do {
            match S.get_init_and_repeat n accu with
            [ Some x -> Left x
            | None -> loop (n + lcm) ]
          }
          else edf_loop tc lcm (time + 1) accu sbl
        }
    }
  | Right msg -> Right msg ]
};

value edf tc stl = do {
  let lcm =
    Sutil.lcm_of_list (List.map (fun st -> Sarray.length st.st_repeat) stl)
  in
  let min_rel_time =
    List.fold_left (fun m st -> min m st.st_release) max_int stl
  in
  let sbl = List.map stask_bucket_of_stask stl in
  match edf_loop tc lcm min_rel_time S.empty sbl with
  [ Left (init, repeat) -> do {
      let contents =
        List.concat (List.map (fun st -> st.st_contents) stl)
      in
      let st =
        {st_release = min_rel_time; st_init = init;
         st_repeat = repeat; st_contents = contents}
      in
let _ = do { eprintf "cnt %d\n" cnt.val; flush stderr } in
      Left st
    }
  | Right msg -> Right msg ]
};

(* response times *)

type resp_time_result =
  [ RTfirst
  | RTlast
  | RTworst ]
;

value response_time_of_task st =
  let a =
    Sarray.append (Sarray.append st.st_init st.st_repeat) st.st_repeat
  in
  fun rtr (itask, td) ->
    loop 0 0 (td.release - st.st_release) where rec loop wrt lrt rel =
      loop_inst rel td.duration where rec loop_inst i dur =
        if dur = 0 then do {
          let wrt = max wrt (i - rel) in
          if rtr = RTfirst then wrt else loop wrt (i - rel) (rel + td.period)
        }
        else if i >= Sarray.length a then
          if rtr = RTlast then lrt else wrt
        else do {
          let dur =
            match Sarray.get a i with
            [ Some tu -> do {
                if tu.tu_itask = itask && not tu.tu_is_preemption then
                  dur - 1
                else dur
              }
            | None -> dur ]
          in
          loop_inst (i + 1) dur
        }
;

value first_response_time st = do {
  let f = response_time_of_task st in
  let wrt =
    List.map (fun (itask, td) -> (itask, f RTfirst (itask, td)))
      st.st_contents
  in
  Array.of_list wrt
};

value last_response_time st = do {
  let f = response_time_of_task st in
  let wrt =
    List.map (fun (itask, td) -> (itask, f RTlast (itask, td)))
      st.st_contents
  in
  Array.of_list wrt
};

value worst_response_time st = do {
  let f = response_time_of_task st in
  let wrt =
    List.map (fun (itask, td) -> (itask, f RTworst (itask, td)))
      st.st_contents
  in
  Array.of_list wrt
};

value global_response_time st = do {
  loop (Sarray.length st.st_repeat - 1) where rec loop i =
    if i < 0 then Sarray.length st.st_init
    else if Sarray.get st.st_repeat i = None then loop (i - 1)
    else Sarray.length st.st_init + i + 1
};
